@@ -1,3 +1,4 @@
+import platform
 import subprocess
 import zlib
 from contextlib import contextmanager
@@ -6,11 +7,12 @@ from math import ceil
 from os import fspath
 from pathlib import Path
 from subprocess import run
-from tempfile import TemporaryDirectory
+from typing import NamedTuple, Sequence
 
 import PIL
 import pytest
-from hypothesis import assume, example, given
+from conftest import needs_python_v
+from hypothesis import assume, given, note, settings
 from hypothesis import strategies as st
 from PIL import Image, ImageChops, ImageCms
 from PIL import features as PIL_features
@@ -29,6 +31,7 @@ from pikepdf import (
     StreamDecodeLevel,
     parse_content_stream,
 )
+from pikepdf.models._transcoding import _next_multiple, unpack_subbyte_pixels
 from pikepdf.models.image import (
     DependencyError,
     NotExtractableError,
@@ -39,6 +42,20 @@ from pikepdf.models.image import (
 # pylint: disable=redefined-outer-name
 
 
+def has_pdfimages():
+    try:
+        run(['pdfimages', '-v'], check=True, capture_output=True)
+    except FileNotFoundError:
+        return False
+    else:
+        return True
+
+
+requires_pdfimages = pytest.mark.skipif(
+    not has_pdfimages(), reason="pdfimages not installed"
+)
+
+
 @pytest.fixture
 def first_image_in(resources, request):
     pdf = None
@@ -80,22 +97,23 @@ def trivial(first_image_in):
 
 @pytest.fixture
 def inline(resources):
-    pdf = Pdf.open(resources / 'image-mono-inline.pdf')
-    for operands, _command in parse_content_stream(pdf.pages[0]):
-        if operands and isinstance(operands[0], PdfInlineImage):
-            yield operands[0], pdf
-    pdf.close()
+    with Pdf.open(resources / 'image-mono-inline.pdf') as pdf:
+        for operands, _command in parse_content_stream(pdf.pages[0]):
+            if operands and isinstance(operands[0], PdfInlineImage):
+                yield operands[0], pdf
+                break
 
 
 def test_image_from_nonimage(resources):
-    pdf = Pdf.open(resources / 'congress.pdf')
-    contents = pdf.pages[0].Contents
-    with pytest.raises(TypeError):
-        PdfImage(contents)
+    with Pdf.open(resources / 'congress.pdf') as pdf:
+        contents = pdf.pages[0].Contents
+        with pytest.raises(TypeError):
+            PdfImage(contents)
 
 
 def test_image(congress):
-    pdfimage = PdfImage(congress[0])
+    xobj, _ = congress
+    pdfimage = PdfImage(xobj)
     pillowimage = pdfimage.as_pil_image()
 
     assert pillowimage.mode == pdfimage.mode
@@ -103,11 +121,13 @@ def test_image(congress):
 
 
 def test_imagemask(congress):
-    assert not PdfImage(congress[0]).image_mask
+    xobj, _ = congress
+    assert not PdfImage(xobj).image_mask
 
 
 def test_imagemask_colorspace(trivial):
-    rawimage = trivial[0]
+    xobj, _ = trivial
+    rawimage = xobj
     rawimage.ImageMask = True
     pdfimage = PdfImage(rawimage)
     assert pdfimage.image_mask
@@ -115,7 +135,8 @@ def test_imagemask_colorspace(trivial):
 
 
 def test_malformed_palette(trivial):
-    rawimage = trivial[0]
+    xobj, _ = trivial
+    rawimage = xobj
     rawimage.ColorSpace = [Name.Indexed, 'foo', 'bar']
     pdfimage = PdfImage(rawimage)
     with pytest.raises(ValueError, match="interpret this palette"):
@@ -123,59 +144,61 @@ def test_malformed_palette(trivial):
 
 
 def test_image_eq(trivial, congress, inline):
+    xobj_trivial, _ = trivial
+    xobj_congress, _ = congress
+    inline_image, _ = inline
     # Note: JPX equality is tested in test_jp2 (if we have a jpeg2000 codec)
-    assert PdfImage(trivial[0]) == PdfImage(trivial[0])
-    assert PdfImage(trivial[0]).__eq__(42) is NotImplemented
-    assert PdfImage(trivial[0]) != PdfImage(congress[0])
+    assert PdfImage(xobj_trivial) == PdfImage(xobj_trivial)
+    assert PdfImage(xobj_trivial).__eq__(42) is NotImplemented
+    assert PdfImage(xobj_trivial) != PdfImage(xobj_congress)
 
-    assert inline != PdfImage(congress[0])
-    assert inline.__eq__(42) is NotImplemented
+    assert inline_image != PdfImage(xobj_congress)
+    assert inline_image.__eq__(42) is NotImplemented
 
 
 def test_image_replace(congress, outdir):
-    pdfimage = PdfImage(congress[0])
+    xobj, pdf = congress
+    pdfimage = PdfImage(xobj)
     pillowimage = pdfimage.as_pil_image()
 
     grayscale = pillowimage.convert('L')
     grayscale = grayscale.resize((4, 4))  # So it is not obnoxious on error
 
-    congress[0].write(zlib.compress(grayscale.tobytes()), filter=Name("/FlateDecode"))
-    congress[0].ColorSpace = Name("/DeviceGray")
-    pdf = congress[1]
+    xobj.write(zlib.compress(grayscale.tobytes()), filter=Name("/FlateDecode"))
+    xobj.ColorSpace = Name("/DeviceGray")
     pdf.save(outdir / 'congress_gray.pdf')
 
 
 def test_lowlevel_jpeg(congress):
-    raw_bytes = congress[0].read_raw_bytes()
+    xobj, _pdf = congress
+    raw_bytes = xobj.read_raw_bytes()
     with pytest.raises(PdfError):
-        congress[0].read_bytes()
+        xobj.read_bytes()
 
     im = Image.open(BytesIO(raw_bytes))
     assert im.format == 'JPEG'
 
-    pim = PdfImage(congress[0])
+    pim = PdfImage(xobj)
     b = BytesIO()
     pim.extract_to(stream=b)
     b.seek(0)
     im = Image.open(b)
-    assert im.size == (congress[0].Width, congress[0].Height)
+    assert im.size == (xobj.Width, xobj.Height)
     assert im.mode == 'RGB'
 
 
 def test_lowlevel_replace_jpeg(congress, outdir):
+    xobj, pdf = congress
     # This test will modify the PDF so needs its own image
-    raw_bytes = congress[0].read_raw_bytes()
+    raw_bytes = xobj.read_raw_bytes()
 
     im = Image.open(BytesIO(raw_bytes))
     grayscale = im.convert('L')
     grayscale = grayscale.resize((4, 4))  # So it is not obnoxious on error
 
-    congress[0].write(
-        zlib.compress(grayscale.tobytes()[:10]), filter=Name("/FlateDecode")
-    )
-    congress[0].ColorSpace = Name('/DeviceGray')
+    xobj.write(zlib.compress(grayscale.tobytes()[:10]), filter=Name("/FlateDecode"))
+    xobj.ColorSpace = Name('/DeviceGray')
 
-    pdf = congress[1]
     pdf.save(outdir / 'congress_gray.pdf')
 
 
@@ -216,55 +239,87 @@ def test_inline_to_pil(inline):
 
 
 def test_bits_per_component_missing(congress):
-    cong_im = congress[0]
+    cong_im, _ = congress
     del cong_im.stream_dict['/BitsPerComponent']
-    assert PdfImage(congress[0]).bits_per_component == 8
+    assert PdfImage(cong_im).bits_per_component == 8
 
 
-@pytest.mark.parametrize(
-    'w,h,pixeldata,cs,bpc',
-    [
-        (1, 1, b'\xff', '/DeviceGray', 1),
-        (1, 1, b'\xf0', '/DeviceGray', 8),
-        (1, 1, b'\xff\x00\xff', '/DeviceRGB', 8),
-        (1, 1, b'\xff\x80\x40\x20', '/DeviceCMYK', 8),
-    ],
-)
-def test_image_roundtrip(outdir, w, h, pixeldata, cs, bpc):
-    pdf = Pdf.new()
+class ImageSpec(NamedTuple):
+    bpc: int
+    width: int
+    height: int
+    colorspace: pikepdf.Name
+    imbytes: bytes
 
-    image_data = pixeldata * (w * h)
 
-    image = Stream(pdf, image_data)
-    image.Type = Name('/XObject')
-    image.Subtype = Name('/Image')
-    image.ColorSpace = Name(cs)
-    image.BitsPerComponent = bpc
-    image.Width = w
-    image.Height = h
+def pdf_from_image_spec(spec: ImageSpec):
+    pdf = pikepdf.new()
+    pdfw, pdfh = 36 * spec.width, 36 * spec.height
 
-    xobj = {'/Im1': image}
-    resources = {'/XObject': xobj}
-    mediabox = [0, 0, 100, 100]
-    stream = b'q 100 0 0 100 0 0 cm /Im1 Do Q'
-    contents = Stream(pdf, stream)
+    pdf.add_blank_page(page_size=(pdfw, pdfh))
 
-    page_dict = {
-        '/Type': Name('/Page'),
-        '/MediaBox': mediabox,
-        '/Contents': contents,
-        '/Resources': resources,
-    }
-    page = pdf.make_indirect(page_dict)
+    imobj = Stream(
+        pdf,
+        spec.imbytes,
+        BitsPerComponent=spec.bpc,
+        ColorSpace=spec.colorspace,
+        Width=spec.width,
+        Height=spec.height,
+        Type=Name.XObject,
+        Subtype=Name.Image,
+    )
 
-    pdf.pages.append(page)
+    pdf.pages[0].Contents = Stream(pdf, b'%f 0 0 %f 0 0 cm /Im0 Do' % (pdfw, pdfh))
+    pdf.pages[0].Resources = Dictionary(XObject=Dictionary(Im0=imobj))
+    pdf.pages[0].MediaBox = Array([0, 0, pdfw, pdfh])
+
+    return pdf
+
+
+@st.composite
+def valid_random_image_spec(
+    draw,
+    bpcs=st.sampled_from([1, 2, 4, 8, 16]),
+    widths=st.integers(min_value=1, max_value=16),
+    heights=st.integers(min_value=1, max_value=16),
+    colorspaces=st.sampled_from([Name.DeviceGray, Name.DeviceRGB, Name.DeviceCMYK]),
+):
+    bpc = draw(bpcs)
+    width = draw(widths)
+    height = draw(heights)
+    colorspace = draw(colorspaces)
+
+    min_imbytes = width * height * (2 if bpc == 16 else 1)
+    if colorspace == Name.DeviceRGB:
+        min_imbytes *= 3
+    elif colorspace == Name.DeviceCMYK:
+        min_imbytes *= 4
+    imbytes = draw(st.binary(min_size=min_imbytes, max_size=2 * min_imbytes))
+
+    return ImageSpec(bpc, width, height, colorspace, imbytes)
+
+
+@given(spec=valid_random_image_spec(bpcs=st.sampled_from([1, 2, 4, 8])))
+@settings(deadline=None)  # For PyPy
+def test_image_save_compare(tmp_path_factory, spec):
+    pdf = pdf_from_image_spec(spec)
+    image = pdf.pages[0].Resources.XObject['/Im0']
+    w = image.Width
+    h = image.Height
+    cs = str(image.ColorSpace)
+    bpc = image.BitsPerComponent
+    pixeldata = image.read_bytes()
+
+    assume((bpc < 8 and cs == '/DeviceGray') or (bpc == 8))
+
+    outdir = tmp_path_factory.mktemp('image_roundtrip')
     outfile = outdir / f'test{w}{h}{cs[1:]}{bpc}.pdf'
     pdf.save(
         outfile, compress_streams=False, stream_decode_level=StreamDecodeLevel.none
     )
 
     with Pdf.open(outfile) as p2:
-        pim = PdfImage(p2.pages[0].Resources.XObject['/Im1'])
+        pim = PdfImage(p2.pages[0].Resources.XObject['/Im0'])
 
         assert pim.bits_per_component == bpc
         assert pim.colorspace == cs
@@ -316,6 +371,130 @@ def test_direct_extract(first_image_in, filename, bpc, filters, ext, mode, forma
     assert im.format == format_
 
 
+def pack_2bit_row(row: Sequence[int]) -> bytes:
+    assert len(row) % 4 == 0
+    im76 = [s << 6 for s in row[0::4]]
+    im54 = [s << 4 for s in row[1::4]]
+    im32 = [s << 2 for s in row[2::4]]
+    im10 = [s << 0 for s in row[3::4]]
+    return bytes(sum(s) for s in zip(im76, im54, im32, im10))
+
+
+def pack_4bit_row(row: Sequence[int]) -> bytes:
+    assert len(row) % 2 == 0
+    upper = [s << 4 for s in row[0::2]]
+    lower = row[1::2]
+    return bytes(sum(s) for s in zip(upper, lower))
+
+
+@st.composite
+def imagelike_data(draw, width, height, bpc, sample_range=None):
+    bits_per_byte = 8 // bpc
+    stride = _next_multiple(width, bits_per_byte)
+
+    if not sample_range:
+        sample_range = (0, 2 ** bpc - 1)
+
+    if bpc in (2, 4, 8):
+        intdata = draw(
+            st.lists(
+                st.lists(
+                    st.integers(*sample_range),
+                    min_size=stride,
+                    max_size=stride,
+                ),
+                min_size=height,
+                max_size=height,
+            )
+        )
+        if bpc == 8:
+            imbytes = b''.join(bytes(row) for row in intdata)
+        elif bpc == 4:
+            imbytes = b''.join(pack_4bit_row(row) for row in intdata)
+        elif bpc == 2:
+            imbytes = b''.join(pack_2bit_row(row) for row in intdata)
+        assert len(imbytes) > 0
+    elif bpc == 1:
+        imdata = draw(
+            st.lists(
+                st.integers(0, 255 if sample_range[1] > 0 else 0),
+                min_size=height * _next_multiple(width, 8),
+                max_size=height * _next_multiple(width, 8),
+            )
+        )
+        imbytes = bytes(imdata)
+    return imbytes
+
+
+class PaletteImageSpec(NamedTuple):
+    bpc: int
+    width: int
+    height: int
+    hival: int
+    colorspace: pikepdf.Name
+    palette: bytes
+    imbytes: bytes
+
+
+def pdf_from_palette_image_spec(spec: PaletteImageSpec):
+    pdf = pikepdf.new()
+    pdfw, pdfh = 36 * spec.width, 36 * spec.height
+
+    pdf.add_blank_page(page_size=(pdfw, pdfh))
+
+    imobj = Stream(
+        pdf,
+        spec.imbytes,
+        BitsPerComponent=spec.bpc,
+        ColorSpace=Array([Name.Indexed, spec.colorspace, spec.hival, spec.palette]),
+        Width=spec.width,
+        Height=spec.height,
+        Type=Name.XObject,
+        Subtype=Name.Image,
+    )
+
+    pdf.pages[0].Contents = Stream(pdf, b'%f 0 0 %f 0 0 cm /Im0 Do' % (pdfw, pdfh))
+    pdf.pages[0].Resources = Dictionary(XObject=Dictionary(Im0=imobj))
+    pdf.pages[0].MediaBox = Array([0, 0, pdfw, pdfh])
+
+    return pdf
+
+
+@st.composite
+def valid_random_palette_image_spec(
+    draw,
+    bpcs=st.sampled_from([1, 2, 4, 8]),
+    widths=st.integers(min_value=1, max_value=16),
+    heights=st.integers(min_value=1, max_value=16),
+    colorspaces=st.sampled_from([Name.DeviceGray, Name.DeviceRGB, Name.DeviceCMYK]),
+    palette=None,
+):
+    bpc = draw(bpcs)
+    width = draw(widths)
+    height = draw(heights)
+    colorspace = draw(colorspaces)
+    hival = draw(st.integers(min_value=0, max_value=(2 ** bpc) - 1))
+
+    imbytes = draw(imagelike_data(width, height, bpc, (0, hival)))
+
+    channels = (
+        1
+        if colorspace == Name.DeviceGray
+        else 3
+        if colorspace == Name.DeviceRGB
+        else 4
+        if colorspace == Name.DeviceCMYK
+        else 0
+    )
+
+    if not palette:
+        palette = draw(
+            st.binary(min_size=channels * (hival + 1), max_size=channels * (hival + 1))
+        )
+
+    return PaletteImageSpec(bpc, width, height, hival, colorspace, palette, imbytes)
+
+
 @pytest.mark.parametrize(
     'filename,bpc,rgb',
     [
@@ -341,6 +520,52 @@ def test_image_palette(resources, filename, bpc, rgb):
     assert im.getpixel((1, 1)) == rgb
 
 
+@contextmanager
+def first_image_from_pdfimages(pdf, tmpdir):
+    if not has_pdfimages():
+        pytest.skip("Need pdfimages for this test")
+
+    pdf.save(tmpdir / 'in.pdf')
+
+    run(
+        ['pdfimages', '-q', '-png', fspath(tmpdir / 'in.pdf'), fspath('pdfimage')],
+        cwd=fspath(tmpdir),
+        check=True,
+    )
+
+    outpng = tmpdir / 'pdfimage-000.png'
+    assert outpng.exists()
+    yield Image.open(outpng)
+
+
+@given(spec=valid_random_palette_image_spec())
+def test_image_palette2(spec, tmp_path_factory):
+    pdf = pdf_from_palette_image_spec(spec)
+    pim = PdfImage(pdf.pages[0].Resources.XObject['/Im0'])
+
+    im1 = pim.as_pil_image()
+
+    with first_image_from_pdfimages(
+        pdf, tmp_path_factory.mktemp('test_image_palette2')
+    ) as im2:
+        if pim.palette.base_colorspace == 'CMYK' and im1.size == im2.size:
+            return  # Good enough - CMYK is hard...
+
+        if im1.mode == im2.mode:
+            diff = ImageChops.difference(im1, im2)
+        else:
+            diff = ImageChops.difference(im1.convert('RGB'), im2.convert('RGB'))
+
+        if diff.getbbox():
+            if pim.palette.base_colorspace in ('L', 'RGB', 'CMYK') and im2.mode == '1':
+                note("pdfimages bug - 1bit image stripped of palette")
+                return
+
+        assert (
+            not diff.getbbox()
+        ), f"{diff.getpixel((0, 0))}, {im1.getpixel((0,0))}, {im2.getpixel((0,0))}"
+
+
 def test_bool_in_inline_image():
     piim = PdfInlineImage(image_data=b'', image_object=(Name.IM, True))
     assert piim.image_mask
@@ -379,8 +604,7 @@ def test_extract_filepath(congress, outdir):
     xobj, _pdf = congress
     pim = PdfImage(xobj)
 
-    # fspath is for Python 3.5
-    result = pim.extract_to(fileprefix=fspath(outdir / 'image'))
+    result = pim.extract_to(fileprefix=(outdir / 'image'))
     assert Path(result).exists()
     assert (outdir / 'image.jpg').exists()
 
@@ -481,6 +705,15 @@ def test_imagemagick_uses_rle_compression(first_image_in):
     assert im.getpixel((5, 5)) == (255, 128, 0)
 
 
+# Unforuntately pytest cannot test for this using "with pytest.warns(...)".
+# Suppression is the best we can manage
+suppress_unraisable_jbigdec_error_warning = pytest.mark.filterwarnings(
+    "ignore:.*jbig2dec error.*:pytest.PytestUnraisableExceptionWarning"
+)
+
+
+@needs_python_v("3.8", reason="for pytest unraisable exception support")
+@suppress_unraisable_jbigdec_error_warning
 def test_jbig2_not_available(jbig2, monkeypatch):
     xobj, _pdf = jbig2
     pim = PdfImage(xobj)
@@ -501,6 +734,12 @@ needs_jbig2dec = pytest.mark.skipif(
 )
 
 
+@needs_jbig2dec
+def test_jbig2_extractor(jbig2):
+    xobj, _pdf = jbig2
+    pikepdf.jbig2.extract_jbig2_bytes(xobj.read_raw_bytes(), b'')
+
+
 @needs_jbig2dec
 def test_jbig2(jbig2):
     xobj, _pdf = jbig2
@@ -531,6 +770,8 @@ def test_jbig2_global_palette(first_image_in):
     assert im.getpixel((0, 0)) == 255  # Ensure loaded
 
 
+@needs_python_v("3.8", reason="for pytest unraisable exception support")
+@suppress_unraisable_jbigdec_error_warning
 def test_jbig2_error(first_image_in, monkeypatch):
     xobj, _pdf = first_image_in('jbig2global.pdf')
     pim = PdfImage(xobj)
@@ -542,10 +783,12 @@ def test_jbig2_error(first_image_in, monkeypatch):
     monkeypatch.setattr(pikepdf.jbig2, 'run', raise_calledprocesserror)
 
     pim = PdfImage(xobj)
-    with pytest.raises(subprocess.CalledProcessError):
+    with pytest.raises(PdfError, match="unfilterable stream"):
         pim.as_pil_image()
 
 
+@needs_python_v("3.8", reason="for pytest unraisable exception support")
+@suppress_unraisable_jbigdec_error_warning
 def test_jbig2_too_old(first_image_in, monkeypatch):
     xobj, _pdf = first_image_in('jbig2global.pdf')
     pim = PdfImage(xobj)
@@ -553,7 +796,9 @@ def test_jbig2_too_old(first_image_in, monkeypatch):
     def run_version_override(subprocargs, *args, **kwargs):
         if '--version' in subprocargs:
             return subprocess.CompletedProcess(subprocargs, 0, 'jbig2dec 0.12\n')
-        return subprocess.run(subprocargs, *args, **kwargs)
+        return subprocess.run(  # pylint: disable=subprocess-run-check
+            subprocargs, *args, **kwargs
+        )
 
     monkeypatch.setattr(pikepdf.jbig2, 'run', run_version_override)
 
@@ -835,93 +1080,46 @@ def test_devicen():
     # pdf.save('devicen.pdf')
 
 
-@pytest.mark.parametrize(
-    'bits, check_pixels',
-    [
-        (2, [(0, 0, 0x00), (1, 1, 0x55)]),
-        (4, [(1, 0, 0xBB), (0, 1, 0x00), (1, 1, 0x11)]),
-    ],
-)
-def test_oddwidth_grayscale(bits, check_pixels):
-    pdf = pikepdf.new()
-    pdf.add_blank_page(page_size=(108, 72))
-
-    imobj = Stream(
-        pdf,
-        bytes([0b00011011, 0b11011000, 0b00000001]),
-        BitsPerComponent=bits,
-        ColorSpace=Name.DeviceGray,
-        Width=3,
-        Height=2,
-        Type=Name.XObject,
-        Subtype=Name.Image,
+@given(
+    spec=valid_random_image_spec(
+        bpcs=st.sampled_from([2, 4]),
+        colorspaces=st.just(Name.DeviceGray),
+        widths=st.integers(1, 7),
+        heights=st.integers(1, 7),
     )
-
-    pdf.pages[0].Contents = Stream(pdf, b'108 0 0 72 0 0 cm /Im0 Do')
-    pdf.pages[0].Resources = Dictionary(XObject=Dictionary(Im0=imobj))
-
+)
+def test_grayscale_stride(spec):
+    pdf = pdf_from_image_spec(spec)
     pim = PdfImage(pdf.pages[0].Resources.XObject.Im0)
     assert pim.mode == 'L'
-    assert pim.bits_per_component == bits
+    imdata = pim.read_bytes()
+    w = pim.width
+    imdata_unpacked_view, stride = unpack_subbyte_pixels(
+        imdata, pim.size, pim.bits_per_component
+    )
+    imdata_unpacked = bytes(imdata_unpacked_view)
+
     bio = BytesIO()
     pim.extract_to(stream=bio)
-    bio.seek(0)
     im = Image.open(bio)
-    assert im.mode == 'L'
-    assert im.size == (3, 2)
+    assert im.mode == 'L' and im.size == pim.size
 
-    # pdf.save(f'oddbit_{bits}.pdf')
-    for check_x, check_y, val in check_pixels:
-        assert im.getpixel((check_x, check_y)) == val
-
-
-def has_pdfimages():
-    try:
-        run(['pdfimages', '-v'], check=True)
-    except FileNotFoundError:
-        return False
-    else:
-        return True
-
-
-requires_pdfimages = pytest.mark.skipif(
-    not has_pdfimages(), reason="pdfimages not installed"
-)
+    for n, pixel in enumerate(im.getdata()):
+        idx = stride * (n // w) + (n % w)
+        assert imdata_unpacked[idx] == pixel
 
 
 @requires_pdfimages
-@given(
-    bpc=st.sampled_from([1, 2, 4, 8, 16]),
-    width=st.integers(min_value=1, max_value=16),
-    height=st.integers(min_value=1, max_value=16),
-    colorspace=st.sampled_from([Name.DeviceGray, Name.DeviceRGB, Name.DeviceCMYK]),
-    imbytes=st.binary(min_size=1, max_size=512),
-)
-def test_random_image(bpc, width, height, colorspace, imbytes, tmp_path_factory):
-
-    pdf = pikepdf.new()
-    pdfw, pdfh = 18 * width, 18 * height
-
-    pdf.add_blank_page(page_size=(pdfw, pdfh))
-    if len(imbytes) < width * height:
-        imbytes = imbytes + bytes(width * height * (2 if bpc == 16 else 1))
-
-    imobj = Stream(
-        pdf,
-        imbytes,
-        BitsPerComponent=bpc,
-        ColorSpace=colorspace,
-        Width=width,
-        Height=height,
-        Type=Name.XObject,
-        Subtype=Name.Image,
-    )
-
-    pdf.pages[0].Contents = Stream(pdf, b'%f 0 0 %f 0 0 cm /Im0 Do' % (pdfw, pdfh))
-    pdf.pages[0].Resources = Dictionary(XObject=Dictionary(Im0=imobj))
-
+@given(spec=valid_random_image_spec())
+def test_random_image(spec, tmp_path_factory):
+    pdf = pdf_from_image_spec(spec)
     pim = PdfImage(pdf.pages[0].Resources.XObject.Im0)
     bio = BytesIO()
+    colorspace = pim.colorspace
+    width = pim.width
+    height = pim.height
+    bpc = pim.bits_per_component
+    imbytes = pim.read_bytes()
     try:
         result_extension = pim.extract_to(stream=bio)
         assert result_extension in ('.png', '.tiff')
@@ -939,11 +1137,11 @@ def test_random_image(bpc, width, height, colorspace, imbytes, tmp_path_factory)
             assert ceil(bpc / 8) * width * height * ncomps > len(imbytes)
             return
         raise
-    except PIL.UnidentifiedImageError as e:
+    except PIL.UnidentifiedImageError:
         if len(imbytes) == 0:
             return
         raise
-    except UnsupportedImageTypeError as e:
+    except UnsupportedImageTypeError:
         if colorspace in (Name.DeviceRGB, Name.DeviceCMYK) and bpc < 8:
             return
         if bpc == 16:
